/* -*- mode: c++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */

/*
 Copyright (C) 2011 Klaus Spanderen

 This program is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the license for more details.

nvcc -I/home/klaus/workspace/QuantLib -I. -arch=compute_20 -O3 -c gpufdmlinearop.cu

*/

#if !defined(QL_NO_UBLAS_SUPPORT)
#include <gpufdmlinearop.hpp>

#include <cusp/csr_matrix.h>
#include <cusp/multiply.h>

namespace QuantLib {

    class GPUFdmLinearOp::pimpl {
      public:
        typedef cusp::csr_matrix<uint32_t, real, cusp::device_memory>
            gpu_matrix_type;

        typedef gpu_matrix_type::index_type   index_type;
        typedef gpu_matrix_type::value_type   value_type;
        typedef gpu_matrix_type::memory_space memory_space;

        typedef cusp::array1d<value_type, memory_space> array1d_type;

        boost::scoped_ptr<array1d_type> x_;
        boost::scoped_ptr<gpu_matrix_type> A_;
    };


    GPUFdmLinearOp::GPUFdmLinearOp(const matrix_type& m)
    : n_      (m.size1()),
      m_      (m.size2()),
      filled1_(m.filled1()),
      filled2_(m.filled2()),
      pImpl_  (new pimpl) {

        pImpl_->A_.reset(new pimpl::gpu_matrix_type(n_, m_, filled2_));

        pImpl_->A_->row_offsets.resize(filled1_);
        thrust::copy(m.index1_data().begin(), m.index1_data().begin()+filled1_,
                     pImpl_->A_->row_offsets.begin());

        pImpl_->A_->column_indices.resize(filled2_);
        thrust::copy(m.index2_data().begin(), m.index2_data().begin()+filled2_,
                     pImpl_->A_->column_indices.begin());

        pImpl_->A_->values.resize(filled2_);
        thrust::copy(m.value_data().begin(), m.value_data().begin()+filled2_,
                     pImpl_->A_->values.begin());

        pImpl_->x_.reset(new pimpl::array1d_type(m_, 0.0));
    }

    GPUFdmLinearOp::~GPUFdmLinearOp() {
        // declare destructor here so that the compiler knows pimpl
    }

    Disposable<GPUFdmLinearOp::array_type>
    GPUFdmLinearOp::apply(const array_type& h_r) const {

        QL_REQUIRE(m_ == h_r.size(), "inconsistent length");

        const pimpl::array1d_type r(h_r.begin(), h_r.end());

        cusp::multiply(*pImpl_->A_, r, *pImpl_->x_);

        const cusp::array1d<pimpl::value_type, cusp::host_memory>
            hv_x = *pImpl_->x_;

        Array retVal(hv_x.begin(), hv_x.end());
        return retVal;
    }


    void GPUFdmLinearOp::apply(real* b, real* x) const {
        const pimpl::array1d_type::view xv(
            thrust::device_ptr<pimpl::value_type>(x),
            thrust::device_ptr<pimpl::value_type>(x+m_));

        const pimpl::array1d_type::view bv(
            thrust::device_ptr<pimpl::value_type>(b),
            thrust::device_ptr<pimpl::value_type>(b+m_));

        cusp::multiply(*pImpl_->A_, bv, xv);
    }


    Disposable<SparseMatrix> GPUFdmLinearOp::toMatrix() const {
        SparseMatrix m(n_, m_, filled2_);

        m.index1_data().resize(filled1_);
        thrust::copy(pImpl_->A_->row_offsets.begin(),
                     pImpl_->A_->row_offsets.begin()+filled1_,
                     m.index1_data().begin());

        m.index2_data().resize(filled2_);
        thrust::copy(pImpl_->A_->column_indices.begin(),
                     pImpl_->A_->column_indices.begin()+filled2_,
                     m.index2_data().begin());

        m.value_data().resize(filled2_);
        thrust::copy(pImpl_->A_->values.begin(),
                     pImpl_->A_->values.begin()+filled2_,
                     m.value_data().begin());

        m.set_filled(filled1_, filled2_);

        return m;
    }
}

#endif
