/* -*- mode: c++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */

/*
 Copyright (C) 2012 Klaus Spanderen

 This program is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the license for more details.

nvcc -I/home/klaus/workspace/QuantLib -I. -arch=compute_20 -O3 -c gpuhundsdorferscheme.cu
*/

/*! \file gpuhundsdorferscheme.cpp
    \brief Hundsdorfer operator splitting
*/


#include <ql/methods/finitedifferences/meshers/fdmmesher.hpp>
#include <ql/methods/finitedifferences/operators/fdmlinearoplayout.hpp>
#include <ql/methods/finitedifferences/operators/fdmlinearopcomposite.hpp>

#include <cusp/array1d.h>
#include <thrust/transform.h>
#include <thrust/functional.h>
#include <thrust/device_ptr.h>

#include <gpuhundsdorferscheme.hpp>

using namespace thrust::placeholders;

namespace QuantLib {
    namespace {
        typedef cusp::array1d<real, cusp::device_memory> array1d_type;
    }


    GPUHundsdorferScheme::GPUHundsdorferScheme(Real theta, Real mu,
        const boost::shared_ptr<FdmMesher>& mesher,
        const boost::shared_ptr<FdmLinearOpComposite>& map)
    : theta_   (theta),
      mu_      (mu),
      n_       (mesher->layout()->size()),
      map_     (map->toMatrix()) {
        const std::vector<SparseMatrix> maps(map->toMatrixDecomp());

        splittings_.reserve(map->size());
        for (Size i=0; i < map->size(); ++i) {
            splittings_.push_back(boost::shared_ptr<GPUTripleBandLinearOp>(
                new GPUTripleBandLinearOp(i, maps[i], mesher)));
        }
    }

    void GPUHundsdorferScheme::setStep(Time dt) {
        dt_ = dt;
    }

    void GPUHundsdorferScheme::step(real* x) {
        const array1d_type::view xv(
            thrust::device_ptr<real>(x), thrust::device_ptr<real>(x+n_));

        array1d_type y(n_);
        map_.apply(x, thrust::raw_pointer_cast(&y[0]));
        cudaThreadSynchronize();

        thrust::transform(y.begin(), y.end(), xv.begin(), y.begin(),
                          _2 + real(dt_)*_1);
        cudaThreadSynchronize();

        array1d_type y0 = y, rhs(n_);

        for (Size i=0; i < splittings_.size(); ++i) {
            splittings_[i]->apply(x, thrust::raw_pointer_cast(&rhs[0]));
            cudaThreadSynchronize();

            thrust::transform(rhs.begin(), rhs.end(), y.begin(), rhs.begin(),
                              real(-theta_*dt_)*_1 + _2);
            cudaThreadSynchronize();

            splittings_[i]->solve_splitting(
                thrust::raw_pointer_cast(&rhs[0]),
                thrust::raw_pointer_cast(&y[0]), -theta_*dt_);
            cudaThreadSynchronize();
        }

        array1d_type yt(n_), ym(n_);
        thrust::transform(y.begin(), y.end(), xv.begin(), ym.begin(),_1 - _2);
        cudaThreadSynchronize();
        map_.apply(thrust::raw_pointer_cast(&ym[0]),
                   thrust::raw_pointer_cast(&yt[0]));
        cudaThreadSynchronize();
        thrust::transform(yt.begin(), yt.end(), y0.begin(), yt.begin(),
                          real(mu_*dt_)*_1+_2);

        for (Size i=0; i < splittings_.size(); ++i) {
            splittings_[i]->apply(thrust::raw_pointer_cast(&y[0]),
                                  thrust::raw_pointer_cast(&rhs[0]));
            cudaThreadSynchronize();
            thrust::transform(rhs.begin(), rhs.end(), yt.begin(), rhs.begin(),
                              real(-theta_*dt_)*_1 + _2);
            cudaThreadSynchronize();
            splittings_[i]->solve_splitting(
                thrust::raw_pointer_cast(&rhs[0]),
                thrust::raw_pointer_cast(&yt[0]), -theta_*dt_);
            cudaThreadSynchronize();
        }

        thrust::copy(yt.begin(), yt.end(), xv.begin());
    }
}
