/* -*- mode: c++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */

/*
 Copyright (C) 2012 Klaus Spanderen

 This program is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the license for more details.

nvcc -I/home/klaus/workspace/QuantLib -I. -arch=compute_20 -O3 -c gputriplebandlinearop.cu
*/

/*! \file gputriplebandlinearop.cu
    \brief C++/Boost/QL interface for cuda sparse matrix
*/

#include <ql/methods/finitedifferences/meshers/fdmmesher.hpp>
#include <ql/methods/finitedifferences/operators/fdmlinearoplayout.hpp>

#include <cusparse_v2.h>
#include <cusp/array1d.h>
#include <thrust/transform.h>
#include <thrust/functional.h>
#include <thrust/iterator/zip_iterator.h>
#include <thrust/iterator/permutation_iterator.h>

#include <gputriplebandlinearop.hpp>

using namespace thrust::placeholders;

namespace QuantLib {
    class GPUTripleBandLinearOp::pimpl {
      public:
        typedef uint32_t            index_type;
        typedef real                value_type;
        typedef cusp::device_memory memory_space;

        typedef cusp::array1d<value_type, memory_space> array1d_type;
        typedef cusp::array1d<value_type, memory_space> index1d_type;

        boost::scoped_ptr<index1d_type> i0_, i2_, revIdx_;
        boost::scoped_ptr<array1d_type> lower_, diag_, upper_;
        boost::scoped_ptr<array1d_type> lowerRevIdx_,diagRevIdx_, upperRevIdx_;
        boost::scoped_ptr<array1d_type> lowerScaled_,diagScaled_, upperScaled_;
        boost::scoped_ptr<array1d_type> rRevIdx_;

        cusparseHandle_t cusparseHandle_;

        pimpl() {
            if (CUSPARSE_STATUS_SUCCESS != cusparseCreate(&cusparseHandle_)) {
                QL_FAIL("can't initialise CUSPARSE");
            }
        }

        void makeRevIdxArray(const array1d_type& in,
                             boost::scoped_ptr<array1d_type>& out) {
            out.reset(new array1d_type(in.size()));

            thrust::copy(thrust::make_permutation_iterator(
                             in.begin(), revIdx_->begin()),
                         thrust::make_permutation_iterator(
                             in.end(), revIdx_->end()),
                         out->begin());
        }
    };


    GPUTripleBandLinearOp::GPUTripleBandLinearOp(
        Size direction,
        const matrix_type& matrix,
        const boost::shared_ptr<FdmMesher>& mesher)
    : n_(mesher->layout()->size()),
      pImpl_(new pimpl),
      lastA_(Null<Real>()) {

        const boost::shared_ptr<FdmLinearOpLayout> layout = mesher->layout();

        std::vector<Size> newDim(layout->dim());
        std::iter_swap(newDim.begin(), newDim.begin()+direction);
        std::vector<Size> newSpacing = FdmLinearOpLayout(newDim).spacing();
        std::iter_swap(newSpacing.begin(), newSpacing.begin()+direction);

        cusp::array1d<pimpl::value_type, cusp::host_memory>
            diag(n_), upper(n_), lower(n_);
        cusp::array1d<pimpl::index_type, cusp::host_memory>
            i0(n_), i2(n_), revIdx(n_);

        const Size dirMax = layout->dim()[direction]-1;

        const FdmLinearOpIterator endIter = layout->end();
        for (FdmLinearOpIterator iter = layout->begin();
            iter!=endIter; ++iter) {

            const Size i = iter.index();
            i0[i] = layout->neighbourhood(iter, direction, -1);
            i2[i] = layout->neighbourhood(iter, direction,  1);

            QL_REQUIRE(   matrix.find_element(i,i)     != 0
                       && matrix.find_element(i,i0[i]) != 0
                       && matrix.find_element(i,i2[i]) != 0,
                          "invalid index access for direction " << direction);

            diag[i]  = matrix(i, i);
            lower[i] = (iter.coordinates()[direction]) 
                       ? matrix(i, i0[i]) : 0;
            upper[i] = (iter.coordinates()[direction] != dirMax) 
                       ? matrix(i, i2[i]) : 0;

            const std::vector<Size>& coordinates = iter.coordinates();
            const Size newIndex =
                  std::inner_product(coordinates.begin(), coordinates.end(),
                                     newSpacing.begin(), Size(0));
            revIdx[newIndex] = i;
        }

        pImpl_->i0_.reset    (new pimpl::index1d_type(i0));
        pImpl_->i2_.reset    (new pimpl::index1d_type(i2));
        pImpl_->diag_.reset  (new pimpl::array1d_type(diag));
        pImpl_->lower_.reset (new pimpl::array1d_type(lower));
        pImpl_->upper_.reset (new pimpl::array1d_type(upper));
        pImpl_->revIdx_.reset(new pimpl::index1d_type(revIdx));
        pImpl_->rRevIdx_.reset(new pimpl::array1d_type(n_));
        pImpl_->diagScaled_.reset (new pimpl::array1d_type(n_));
        pImpl_->lowerScaled_.reset(new pimpl::array1d_type(n_));
        pImpl_->upperScaled_.reset(new pimpl::array1d_type(n_));

        pImpl_->makeRevIdxArray(*pImpl_->lower_, pImpl_->lowerRevIdx_);
        pImpl_->makeRevIdxArray(*pImpl_->upper_, pImpl_->upperRevIdx_);
        pImpl_->makeRevIdxArray(*pImpl_->diag_, pImpl_->diagRevIdx_);
        cudaDeviceSynchronize();
    }

    GPUTripleBandLinearOp::~GPUTripleBandLinearOp() {
        // declare destructor here so that the compiler knows pimpl
    }

    Disposable<SparseMatrix> GPUTripleBandLinearOp::toMatrix() const {
        QL_FAIL("not implemented");
    }

    namespace {
        typedef thrust::tuple<real,real,real> real3;

        struct TripleProduct
            : public thrust::binary_function<real3,real3,real> {

            __host__ __device__
            real operator()(const real3& a, const real3& b) const {
                return thrust::get<0>(a) * thrust::get<0>(b) +
                       thrust::get<1>(a) * thrust::get<1>(b) +
                       thrust::get<2>(a) * thrust::get<2>(b);
            }
        };
    }

    void GPUTripleBandLinearOp::apply(real* x, real* y) const {
        const pimpl::array1d_type::view xv(
            thrust::device_ptr<pimpl::value_type>(x),
            thrust::device_ptr<pimpl::value_type>(x+n_));

        const pimpl::array1d_type::view yv(
            thrust::device_ptr<pimpl::value_type>(y),
            thrust::device_ptr<pimpl::value_type>(y+n_));

        thrust::transform(
            thrust::make_zip_iterator(
                thrust::make_tuple(pImpl_->lower_->begin(),
                                   pImpl_->diag_->begin(),
                                   pImpl_->upper_->begin())),
            thrust::make_zip_iterator(
                thrust::make_tuple(pImpl_->lower_->end(),
                                   pImpl_->diag_->end(),
                                   pImpl_->upper_->end())),
            thrust::make_zip_iterator(
                thrust::make_tuple(
                    thrust::make_permutation_iterator(
                        xv.begin(), pImpl_->i0_->begin()),
                    xv.begin(),
                    thrust::make_permutation_iterator(
                        xv.begin(), pImpl_->i2_->begin()))),
            yv.begin(),
            TripleProduct());
    }

    Disposable<GPUTripleBandLinearOp::array_type>
        GPUTripleBandLinearOp::apply(const array_type& h_r) const {

        QL_REQUIRE(n_ == h_r.size(), "inconsistent length");
        pimpl::array1d_type r(h_r.begin(), h_r.end());

        pimpl::array1d_type y(n_);
        apply(thrust::raw_pointer_cast(r.data()),
              thrust::raw_pointer_cast(y.data()));

        const cusp::array1d<pimpl::value_type, cusp::host_memory> hv_y = y;

        Array retVal(hv_y.begin(), hv_y.end());
        return retVal;
    }

    void GPUTripleBandLinearOp::cusparseXgtsc(float* l, float* d, 
                                              float* u, float* x) const {
        if (CUSPARSE_STATUS_SUCCESS !=
            cusparseSgtsv(pImpl_->cusparseHandle_, n_, 1, l, d, u, x, n_)) {
            QL_FAIL("sgtsv failed");
        }
        cudaDeviceSynchronize();
    }

    void GPUTripleBandLinearOp::cusparseXgtsc(double* l, double* d, 
                                              double* u, double* x) const {
        if (CUSPARSE_STATUS_SUCCESS !=
            cusparseDgtsv(pImpl_->cusparseHandle_, n_, 1, l, d, u, x, n_)) {
            QL_FAIL("sgtsv failed");
        }
        cudaDeviceSynchronize();
    }

    void GPUTripleBandLinearOp::solve_splitting(real* r, real* x, Real a) 
    const {
        const pimpl::array1d_type::view xv(
            thrust::device_ptr<pimpl::value_type>(x),
            thrust::device_ptr<pimpl::value_type>(x+n_));

        const pimpl::array1d_type::view rv(
            thrust::device_ptr<pimpl::value_type>(r),
            thrust::device_ptr<pimpl::value_type>(r+n_));

        if (lastA_ != a) {
            thrust::transform(pImpl_->lowerRevIdx_->begin(),
                              pImpl_->lowerRevIdx_->end(),
                              pImpl_->lowerScaled_->begin(), a*_1);
            thrust::transform(pImpl_->upperRevIdx_->begin(),
                              pImpl_->upperRevIdx_->end(),
                              pImpl_->upperScaled_->begin(), a*_1);
            thrust::transform(pImpl_->diagRevIdx_->begin(),
                              pImpl_->diagRevIdx_->end(),
                              pImpl_->diagScaled_->begin(),  a*_1+real(1));
            lastA_ = a;
        }
        thrust::copy(thrust::make_permutation_iterator(
                         rv.begin(), pImpl_->revIdx_->begin()),
                     thrust::make_permutation_iterator(
                         rv.end(), pImpl_->revIdx_->end()),
                     pImpl_->rRevIdx_->begin());
        cudaDeviceSynchronize();

        cusparseXgtsc(
                thrust::raw_pointer_cast(&(*pImpl_->lowerScaled_)[0]),
                thrust::raw_pointer_cast(&(*pImpl_->diagScaled_)[0]),
                thrust::raw_pointer_cast(&(*pImpl_->upperScaled_)[0]),
                thrust::raw_pointer_cast(&(*pImpl_->rRevIdx_)[0]));

        thrust::copy(pImpl_->rRevIdx_->begin(), pImpl_->rRevIdx_->end(),
                     thrust::make_permutation_iterator(
                          xv.begin(), pImpl_->revIdx_->begin()));
    }


    Disposable<GPUTripleBandLinearOp::array_type>
        GPUTripleBandLinearOp::solve_splitting(const array_type& h_r,
                                               Real a) {

        QL_REQUIRE(n_ == h_r.size(), "inconsistent length");

        pimpl::array1d_type x(n_);
        pimpl::array1d_type r(h_r.begin(), h_r.end());

        solve_splitting(thrust::raw_pointer_cast(&r[0]),
                        thrust::raw_pointer_cast(&x[0]), a);

        const cusp::array1d<pimpl::value_type, cusp::host_memory> hv_x = x;
        Array retVal(hv_x.begin(), hv_x.end());

        return retVal;
    }
}
