/* -*- mode: c++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */

/*
 Copyright (C) 2012 Klaus Spanderen
 This program is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the license for more details.

g++ -Wall -I. -I/usr/local/cuda/include -I/home/klaus/workspace/QuantLib -o hm hm.cpp cudautils.o gpuhundsdorferscheme.o gputriplebandlinearop.o gpufdmlinearop.o -L/home/klaus/workspace/QuantLib/ql/.libs -lQuantLib -lboost_chrono -lboost_system -L/usr/local/cuda/lib64 -lcudart -lcusparse
 */

#include <ql/quotes/simplequote.hpp>
#include <ql/time/calendars/nullcalendar.hpp>
#include <ql/time/daycounters/actual365fixed.hpp>
#include <ql/instruments/payoffs.hpp>
#include <ql/instruments/europeanoption.hpp>
#include <ql/models/equity/hestonmodel.hpp>
#include <ql/processes/ornsteinuhlenbeckprocess.hpp>
#include <ql/termstructures/yield/flatforward.hpp>
#include <ql/methods/finitedifferences/operators/fdmhestonop.hpp>
#include <ql/methods/finitedifferences/operators/fdmlinearoplayout.hpp>
#include <ql/methods/finitedifferences/schemes/hundsdorferscheme.hpp>
#include <ql/methods/finitedifferences/meshers/fdmmeshercomposite.hpp>
#include <ql/methods/finitedifferences/meshers/fdmhestonvariancemesher.hpp>
#include <ql/methods/finitedifferences/meshers/fdmblackscholesmesher.hpp>

#include <gpufdmlinearop.hpp>
#include <gputriplebandlinearop.hpp>
#include <gpuhundsdorferscheme.hpp>

#include <boost/chrono.hpp>
#include <boost/assign/list_of.hpp>

#include <thrust/host_vector.h>
#include <thrust/device_vector.h>

#include <cuda_runtime.h>
#include <stdio.h>

using namespace QuantLib;

namespace {
    boost::shared_ptr<YieldTermStructure>
    flatRate(Rate forward, const DayCounter& dc) {
        return boost::shared_ptr<YieldTermStructure>(
            new FlatForward(0, NullCalendar(), 
                Handle<Quote>(boost::shared_ptr<Quote>(
                    new SimpleQuote(forward))), dc));
    }

    struct FdHestonGrid {
        const char* const name;
        Size t, x, v;
    };

    struct HestonModelData {
        const char* const name;
        Real v0;
        Real kappa;
        Real theta;
        Real sigma;
        Real rho;
        Real r;
        Real q;
    };

    HestonModelData hestonModels[] = {
        // ADI finite difference schemes for option pricing in the
        // Heston model with correlation, K.J. in t'Hout and S. Foulon,
        {"'t Hout case 1", 0.04, 1.5, 0.04, 0.3, -0.9, 0.025, 0.0},
        {"'t Hout case 2", 0.12, 3.0, 0.12, 0.04, 0.6, 0.01, 0.04},
        {"'t Hout case 3", 0.0707,0.6067, 0.0707, 0.2928, -0.7571, 0.03, 0.0},
        {"'t Hout case 4", 0.06, 2.5, 0.06, 0.5, -0.1, 0.0507, 0.0469},
        // Efficient numerical methods for pricing American options under
        // stochastic volatility, Samuli Ikonen and Jari Toivanen,
        {"Ikonen-Toivanen", 0.0625, 5, 0.16, 0.9, 0.1, 0.1, 0.0},
        // Not-so-complex logarithms in the Heston model,
        // Christian Kahl and Peter Jäckel
        {"Kahl-Jaeckel", 0.16, 1.0, 0.16, 2.0, -0.8, 0.0, 0.0},
        // self defined test cases
        {"Equity case", 0.07, 2.0, 0.04, 0.55, -0.8, 0.03, 0.00 },
        {"high correlation", 0.07, 1.0, 0.04, 0.55,  0.995, 0.02, 0.04 },
        {"low Vol-Of-Vol", 0.07, 1.0, 0.04, 0.001, -0.75, 0.04, 0.03 },
        {"kappaEqSigRho", 0.07, 0.4, 0.04, 0.5, 0.8, 0.03, 0.03 }
    };

    boost::shared_ptr<HestonProcess> makeHestonProcess(
                                             const HestonModelData& params) {

        Handle<Quote> spot(
                boost::shared_ptr<SimpleQuote>(new SimpleQuote(100)));

        DayCounter dayCounter = Actual365Fixed();
        Handle<YieldTermStructure> rTS(flatRate(params.r, dayCounter));
        Handle<YieldTermStructure> qTS(flatRate(params.q, dayCounter));

        return boost::shared_ptr<HestonProcess>(new HestonProcess(
                   rTS, qTS, spot, params.v0, params.kappa,
                   params.theta, params.sigma, params.rho));
    }
}

void testGPUHundsdorferScheme() {
    const Time maturity = 1.0;
    const Date maturityDate = Date(Settings::instance().evaluationDate())
        + Period(Size(maturity*365), Days);

    const boost::shared_ptr<Exercise> exercise(
        new EuropeanExercise(maturityDate));
    const boost::shared_ptr<StrikedTypePayoff> payoff(
        new PlainVanillaPayoff(Option::Call, 100));

    const boost::shared_ptr<VanillaOption> option(
        new VanillaOption(payoff, exercise));

    FdHestonGrid grids[] = { { "50x200x100",    50,  200,  100 },
                             { "100x200x100",  100,  200,  100 },
                             { "100x500x100",  100,  500,  100 },
                             { "100x500x200",  100,  500,  200 },
                             { "100x1000x500", 100, 1000,  500 },
                             { "100x2000x500", 100, 2000,  500 },
                             { "100x2000x1000",100, 2000, 1000 } };

    const HestonModelData hestonModelParams = hestonModels[6]; // equity case

    const boost::shared_ptr<HestonProcess> hestonProcess(
        makeHestonProcess(hestonModelParams));
    const boost::shared_ptr<HestonModel> hestonModel(
        new HestonModel(hestonProcess));

    for (Size i=0; i < sizeof(grids)/sizeof(FdHestonGrid); ++i) {
        const Size tGrid = grids[i].t;
        const Size xGrid = grids[i].x;
        const Size vGrid = grids[i].v;

        const boost::shared_ptr<FdmHestonVarianceMesher> varianceMesher(
            new FdmHestonVarianceMesher(vGrid, hestonProcess, maturity, 5));

        const boost::shared_ptr<Fdm1dMesher> equityMesher(
            new FdmBlackScholesMesher(
                xGrid, FdmBlackScholesMesher::processHelper(
                    hestonProcess->s0(), hestonProcess->dividendYield(),
                    hestonProcess->riskFreeRate(),
                    varianceMesher->volaEstimate()),
                    maturity, payoff->strike(),
                    Null<Real>(), Null<Real>(), 0.0001, 1.5,
                    std::pair<Real, Real>(payoff->strike(), 0.1)));

        const boost::shared_ptr<FdmMesher> mesher(
            new FdmMesherComposite(equityMesher, varianceMesher));

        const boost::shared_ptr<FdmLinearOpLayout> layout(mesher->layout());
        Array x(mesher->layout()->size());
        for (FdmLinearOpIterator iter = layout->begin();
             iter != layout->end(); ++iter) {
            x[iter.index()] = (*payoff)(std::exp(mesher->location(iter, 0)));
        }

        thrust::device_vector<real> d_x(x.begin(), x.end());

        const Real dt = maturity/tGrid;
        const boost::shared_ptr<FdmLinearOpComposite> map(
            new FdmHestonOp(mesher, hestonProcess));
        map->setTime(0, dt);

        const Real mu = 0.5;
        const Real theta = 1.0-std::sqrt(2.0)/2;
        
        const boost::shared_ptr<HundsdorferScheme> scheme(
            new HundsdorferScheme(theta, mu, map));
        scheme->setStep(dt);
        const boost::shared_ptr<GPUHundsdorferScheme> gpuScheme(
            new GPUHundsdorferScheme(theta, mu, mesher, map));
        gpuScheme->setStep(dt);
        
        boost::chrono::high_resolution_clock::time_point t0
            = boost::chrono::high_resolution_clock::now();

        const std::vector<Size> corr 
            = boost::assign::list_of(xGrid/2)(vGrid/2);

        for (Real t = maturity; t > 0; t-=dt) {
            scheme->step(x, t);
        }
        const Real cpuPrice = x[layout->index(corr)];
        const Time cpuTime = boost::chrono::duration<double>(
            boost::chrono::high_resolution_clock::now()-t0).count();


        t0= boost::chrono::high_resolution_clock::now();

        for (Real t = maturity; t > 0; t-=dt) {
            gpuScheme->step(thrust::raw_pointer_cast(&d_x[0]));
            cudaThreadSynchronize();
        }

        const Real gpuPrice = d_x[layout->index(corr)];
        const Time gpuTime = boost::chrono::duration<double>(
            boost::chrono::high_resolution_clock::now()-t0).count();

        printf("%s :\tdiff: %g\t cpu: %f\t gpu: %f\t spd-up: %f\n",
               grids[i].name, (cpuPrice-gpuPrice)/cpuPrice, 
               cpuTime, gpuTime, cpuTime/gpuTime);
    }
}

int main() {
    testGPUHundsdorferScheme();
    return 0;
}
